// July 6th, 2009
// WinDBG extension to read BigLib arbitrary precision numbers

#include <windows.h>
#include <dbgeng.h>

void DllMain()
{
}

extern "C" HRESULT __stdcall DebugExtensionInitialize(PULONG Version, PULONG Flags)
{
    HRESULT hFuncRet=E_FAIL;

    PDEBUG_CONTROL4 control=0;
    PDEBUG_CLIENT4 client=0;
    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;

    control->Output(DEBUG_OUTPUT_NORMAL,"extension: initializing...\n");

    *Version = DEBUG_EXTENSION_VERSION(1, 0);
    *Flags = 0;

    hFuncRet = S_OK;

    cleanup:

    if(control)
        control->Release();
    if(client)
        client->Release();

    return hFuncRet;
}

extern "C" VOID __stdcall DebugExtensionUninitialize(void)
{
    PDEBUG_CONTROL4 control=0;
    PDEBUG_CLIENT4 client=0;
    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;

    control->Output(DEBUG_OUTPUT_NORMAL,"extension: uninitializing...\n");

    cleanup:

    if(control)
        control->Release();
    if(client)
        client->Release();
}

extern "C" VOID WINAPI DebugExtensionNotify(ULONG Notify, ULONG64 Argument)
{
    PDEBUG_CONTROL4 control=0;
    PDEBUG_CLIENT4 client=0;
    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;

    switch(Notify)
    {
        case DEBUG_NOTIFY_SESSION_ACTIVE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_ACTIVE\n");
            break;

        case DEBUG_NOTIFY_SESSION_INACTIVE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_INACTIVE\n");
            break;

        case DEBUG_NOTIFY_SESSION_ACCESSIBLE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_ACCESSIBLE\n");
            break;

        case DEBUG_NOTIFY_SESSION_INACCESSIBLE:
            //control->Output(DEBUG_OUTPUT_NORMAL,"extension: got DEBUG_NOTIFY_SESSION_INACCESSIBLE\n");
            break;
    }

    cleanup:

    if(control)
        control->Release();
    if(client)
        client->Release();
}

HRESULT __stdcall bn(PDEBUG_CLIENT4 Client, PCSTR arg)
{
    HRESULT hRes = E_FAIL;

    UINT64 pBig = 0;

    UINT n_digits = 0;
    PDWORD digits = 0;

    DWORD dwRet = 0;

    // get interfaces
    //
    PDEBUG_CONTROL4 control = 0;
    PDEBUG_CLIENT4 client = 0;
    PDEBUG_DATA_SPACES3 dataspaces = 0;

    if(DebugCreate(__uuidof(IDebugClient4), (PVOID *)&client) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugControl3), (PVOID *)&control) != S_OK)
        goto cleanup;
    if(client->QueryInterface(__uuidof(IDebugDataSpaces3), (PVOID *)&dataspaces) != S_OK)
        goto cleanup;

    // get arg
    //
    DEBUG_VALUE pdv;

    memset(&pdv,0,sizeof(pdv));

    if(S_OK != control->Evaluate(arg, DEBUG_VALUE_INT64, &pdv, 0))
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: evaluating argument\n");
        goto cleanup;
    }

    pBig = pdv.I64;

    control->Output(DEBUG_OUTPUT_NORMAL, "bignum @%08X\n", pBig);

    // roy's big numbers appear to look like this:
    //
    // struct
    // {
    //     DWORD n_digits;
    //     DWORD digits[1]; // [d<0>, d<1>, d<n_digits-1>] where d0 is the least significant
    // };

    // get n_digits member
    //
    if(dataspaces->ReadVirtual(pBig, &n_digits, 4, &dwRet) != S_OK)
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: reading memory form target\n");
        goto cleanup;
    }

    // sanity check, allocate spaces for digits
    //
    if(n_digits > 1024)
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: more than 1024 digits, unlikely this is a real pointer to a bignum, bailing!\n");
        goto cleanup;
    }

    if(n_digits == 0)
    {
        control->Output(DEBUG_OUTPUT_NORMAL, "0\n");
        goto cleanup;
    }

    // read in digits
    //
    digits = new DWORD[n_digits * sizeof(DWORD)];
    
    if(dataspaces->ReadVirtual(pBig + 4, digits, n_digits * sizeof(DWORD), &dwRet) != S_OK)
    {
        control->Output(DEBUG_OUTPUT_ERROR, "ERROR: reading memory form target\n");
        goto cleanup;
    }    
    
    // print 'em
    //
    for(INT i=n_digits-1; i>=0; --i)
        control->Output(DEBUG_OUTPUT_NORMAL, "%08X", digits[i]);

    control->Output(DEBUG_OUTPUT_NORMAL, "\n");

    // done!
    //
    hRes = S_OK;

    cleanup:

    if(dataspaces)
        dataspaces->Release();
    if(control)
        control->Release();
    if(client)
        client->Release();
    if(digits)
        delete [] digits;

    return hRes;
}

